# ######################################################################### #
#  dll_loader.py	--	Python module wrapping AVMUDLL/VNADLL				#
# 																			#
#  	Adapted from /libbytecode/vnadll/akela_vna_dll.h							#
# 																			#
# 	Author: Connor Wolf <cwolf@akelainc.com>								#
# 																			#
# ######################################################################### #


import os.path
import platform
import sys
import traceback
import shutil


def get_search_paths():
	''' Build a list of search paths where we should look for the
	AVMU DLL.

	'''
	locations = []
	loc = os.path.dirname(os.path.abspath(__file__))
	up = os.path.abspath(os.path.join(loc, "../"))
	upp = os.path.abspath(os.path.join(up, "../"))
	build_dir_dbg1 = os.path.abspath(os.path.join(loc, "../../../x64/Debug/"))
	build_dir_dbg2 = os.path.abspath(os.path.join(loc, "../../../Out/x64/Debug/"))
	build_dir_rls1 = os.path.abspath(os.path.join(loc, "../../../Out/x64/Release/"))
	build_dir_rls2 = os.path.abspath(os.path.join(loc, "../../../Out/x64/ReleaseWPrivateApi/"))
	build_dir_rls3 = os.path.abspath(os.path.join(loc, "../../../../../Out/x64/ReleaseWPrivateApi/"))
	build_dir_rls4 = os.path.abspath(os.path.join(loc, "../../../../../Out/x64/Release/"))
	locations.append(build_dir_dbg1)
	locations.append(build_dir_dbg2)
	locations.append(build_dir_rls1)
	locations.append(build_dir_rls2)
	locations.append(build_dir_rls3)
	locations.append(build_dir_rls4)
	locations.append(loc)
	locations.append(upp)
	locations.append(up)



	split_on = {"Linux" : ":", "Windows" : ";"}
	split = os.environ['PATH'].split(split_on[platform.system()])

	if getattr(sys, 'frozen', False):
		# we are running in a |PyInstaller| bundle
		locations.append(sys._MEIPASS)

	# Validate and canonize the various paths
	split = [os.path.abspath(item) for item in split if os.path.exists(item)]

	ret = locations+split
	return ret

def check_copy_to_local(fq_dll_path, dll_name):
	lib_dir = os.path.dirname(os.path.abspath(__file__))

	# Dll is in the library directory. Just use it as-is.
	if fq_dll_path.startswith(lib_dir):
		return fq_dll_path

	to_path = os.path.join(lib_dir, dll_name)


	if os.path.exists(to_path):
		with open(fq_dll_path, "rb") as f_a, open(to_path, "rb") as f_b:
			if f_a.read() == f_b.read():
				return to_path
			else:
				print("Dll in library directory differs from build directory! Attempting to copy it there")
	else:
		print("DLL is not located in the library directory. Attempting to copy it there")

	try:
		shutil.copy(fq_dll_path, to_path)
		print("Library copied into tree. Path: %s" % (to_path))
		return to_path
	except Exception:
		print("Failed to move dll into library directory!")
		print("Target location: %s" % (to_path, ))
		print("DLL Location: %s" % (fq_dll_path, ))
		traceback.print_exc()
		return fq_dll_path

def find_dll():
	''' Search both the local working directory, and the
	system environment (`PATH`) for the DLL/SO.
	'''

	dll_lut = {"Linux" : "libavmu.so", "Windows" : "avmudll.dll"}

	plat = platform.system()

	if plat in dll_lut:
		dll_name = dll_lut[plat]
	else:
		raise RuntimeError("Unknown platform: '%s'" % platform.system())

	locations = get_search_paths()


	for location in locations:
		fq_dll_path = os.path.join(location, dll_name)
		if os.path.exists(fq_dll_path):
			print("Found avmu dll at path: '{}'".format(fq_dll_path))
			return check_copy_to_local(fq_dll_path, dll_name)
	raise ValueError("Could not find DLL/SO! Searched paths: \n	- %s" % "\n	- ".join(locations))


STATIC_FFI = None
STATIC_LIB = None

def load_ffi_interface():
	'''
	Load and return the FFI library instance, and DLL interface to the avmu DLL.

	return value is a 2-tuple (ffi_lib, dll_handle)
	'''

	# We have to use globals or some sort of persistent state here, because if
	# we load the DLL twice, it'll create two typedef classes for all the DLL objects,
	# which will then fail to interoperate.
	global STATIC_FFI
	global STATIC_LIB
	if STATIC_FFI is not None and STATIC_LIB is not None:
		return STATIC_FFI, STATIC_LIB

	dll_path = find_dll()

	from . import load_header
	from cffi import FFI


	ffi = FFI()
	headers = load_header.load()

	ffi.cdef(headers)
	lib = ffi.dlopen(dll_path)

	print("Loaded library version: ", ffi.string(lib.versionString()).decode("utf-8"))

	STATIC_FFI = ffi
	STATIC_LIB = lib

	return ffi, lib





import tqdm
import random
from . import avmu_library
from . import avmu_exceptions

def is_avmu_alive(target_ip, port=None):
	'''
	Context-free "Is there something at this IP" command.

	Uses a random port between 1050 and 1279, so multiple calls at the same time
	*should* be generally harmless, though if you're pinging the same unit the
	AVMU may get confused, and some of the ping responses may get lost.

	Functionally, the AVMU sends it's responses to the last IP address it received
	data from. Therefore, arbitrary UDP traffic can cause the AVMU to change where
	it's sending data.
	'''

	if port is None:
		port = random.randint(1050, 1279)

	device = avmu_library.AvmuInterface()
	device.setIPAddress(target_ip)
	device.setIPPort(port)
	device.setTimeout(100)

	try:
		device.utilPingUnit()
		return True
	except avmu_exceptions.Avmu_Exception:
		return False

def get_avmu_info(target_ip):
	'''
	Given an AVMU at a specified address, connect to it, and
	return it's hardware details.
	'''

	device = avmu_library.AvmuInterface()
	device.setIPAddress(target_ip)
	device.setIPPort(random.randint(1050, 1279))
	device.setTimeout(100)
	device.initialize()

	deets = device.getHardwareDetails()
	return deets


if __name__ == '__main__':
	import logging
	import sys
	import pprint
	# logging.basicConfig(level=logging.DEBUG)


	if len(sys.argv) >= 2:
		tgt = sys.argv[1]
		info = get_avmu_info(tgt)
		print("Avmu info: ")
		pprint.pprint(info)
	else:
		print("No Arguments!")






import sys

from . import dll_loader
from . import avmu_library

def header_update():

	ffi, dll = dll_loader.load_ffi_interface()


def dispatch(mode, args):
	funcs = {
		'gen_headers' : header_update
	}

	if mode in funcs:
		funcs[mode](*args)
	else:
		print("Error! No function matching '%s'!" % (mode))

def usage():
	print("Avmu testing widget")
	print("Usage: python(3) -m avmu <mode> [args]")
	print("")
	print("'Modes:")
	print("	gen_headers	- Regenerate headers from the library sources (Only useful for development)")

def go():
	print("AVMU CLI Test")
	if len(sys.argv) == 1:
		usage()
		return
	if len(sys.argv) > 1:
		mode, args = sys.argv[1], sys.argv[2:]
		dispatch(mode, args)


if __name__ == '__main__':
	go()






import os.path
import sys
import re

# THIS IS HORRRRRRRIIIIBBBBBBLLLEEEEE
# It works, though. -cw
HEADER_ROOT = os.path.abspath(os.path.join(__file__, "../../../../dlls/"))

AVMU_HEADER = os.path.join(HEADER_ROOT, "avmudll", "akela_avmu_dll.h")
VNA_HEADER   = os.path.join(HEADER_ROOT, "vnadll", "akela_vna_dll.h")
COMM_HEADER  = os.path.join(HEADER_ROOT, "dll_common", "dll_common.h")

def replace_preprocessor(in_header, no_declspec, cpp=False, htype=None):
	'''
	Join and rewrite the application headers so they're in a format that cffi can
	handle.

	This mostly involves a BUNCH of REALLY horrible string parsing hackery.
	'''
	conditional_split = re.compile(r"// <<<<<< CONDITIONAL START \(do not change this line\!\)|// \(do not change this line\!\) CONDITIONAL END >>>>>>>>")
	if conditional_split.search(in_header):
		pre, dll_def, post = re.split(r"// <<<<<< CONDITIONAL START \(do not change this line\!\)|// \(do not change this line\!\) CONDITIONAL END >>>>>>>>", in_header)
		in_header = pre + post
		dll_def = ""

	in_header = in_header.replace("CONDITIONAL_EXTERN", "")

	if htype == "avmu":
		in_header = in_header.replace("xxxDLL_API",   "AVMUDLL_API")
	elif htype == "vna":
		in_header = in_header.replace("xxxDLL_API",   "VNADLL_API")
		pass
	elif htype is None:
		pass
	else:
		raise ValueError("Invalid header type: '%s'" % htype)

	if no_declspec:
		print("Doing no __declspec generation")
		in_header = in_header.replace("AVMUDLL_API", "__declspec(dllimport)")
		in_header = in_header.replace("VNADLL_API",   "__declspec(dllimport)")
		in_header = in_header.replace("xxxDLL_API",   "__declspec(dllimport)")
		in_header = in_header.replace("__declspec(dllimport)", "")
	else:
		print("Doing __declspec generation")
		in_header = in_header.replace("AVMUDLL_API", "__declspec(dllimport)")
		in_header = in_header.replace("VNADLL_API",   "__declspec(dllimport)")
		in_header = in_header.replace("xxxDLL_API",   "__declspec(dllimport)")

	if not cpp:
		in_header = re.sub("// <<<<<< CPP WRAP START.*?CPP WRAP END >>>>>>>>", "", in_header, flags=re.DOTALL)
		in_header = in_header.replace(" = 5", "")    # Default arguments break the parser
		in_header = in_header.replace(" = 0", "")    # Default arguments break the parser
		in_header = in_header.replace(" = true", "") # Default arguments break the parser
		in_header = in_header.replace(" = NULL", "") # Default arguments break the parser

	in_header = re.sub("// <<<<<< SNIP START.*?SNIP END >>>>>>>>", "", in_header, flags=re.DOTALL)

	lines = []
	if not cpp:
		for line in in_header.split("\n"):
			if line.strip().startswith("#"):
				continue
			lines.append(line)

		in_header = "\n".join(lines)

	return in_header, dll_def

def assemble_header(main_header, common_header, output_path, no_declspec=True, cpp=False, header_name="__AKELA_DLL_HEADER", htype=None):
	mh = open(main_header).read()
	ch = open(common_header).read()

	# print(mh)
	# print(ch)

	mh, dll_def   = replace_preprocessor(mh, no_declspec, cpp, htype)
	ch, dummy_def = replace_preprocessor(ch, no_declspec, cpp, htype)

	pre, dummy, post = re.split("// <<<<<< INCLUDE START|INCLUDE END >>>>>>>>", mh)

	agg_h = pre + ch + post
	if not no_declspec:
		agg_h = """
#ifndef {hname}
#define {hname}
		""".format(hname=header_name) + agg_h + """
#endif

		"""

	splice_key = r"// <<<<<< SPLICE_POINT (do not change this line!) >>>>>>"
	if splice_key in agg_h and not no_declspec:
		# print("")
		# print("")
		# print("")
		# print("Splice point!")
		# print(dll_def)
		# print("")
		# print("")
		# print("")
		agg_h = agg_h.replace(splice_key, dll_def)

	dpath = os.path.split(output_path)[0]
	if not os.path.exists(dpath):
		os.makedirs(dpath)
	print("Writing to filepath: ", output_path)
	with open(output_path, "w") as fp:
		fp.write(agg_h)
	return agg_h



def load():
	header_root = os.path.join(os.path.dirname(__file__), "headers")

	try:
		agg_avmu = assemble_header(
				main_header   = AVMU_HEADER,
				common_header = COMM_HEADER,
				output_path   = os.path.join(header_root, "avmu_header_agg.h")
			)
		_        = assemble_header(
				main_header   = AVMU_HEADER,
				common_header = COMM_HEADER,
				output_path   = os.path.join(header_root, "avmu_header_agg_c.h"),
				no_declspec   = False,
				cpp           = True,
				header_name   = "__AKELA_AVMU_DLL_HEADER", htype='avmu'
			)
		_        = assemble_header(
				main_header   = VNA_HEADER,
				common_header = COMM_HEADER,
				output_path   = os.path.join(header_root, "vna_header_agg_c.h"),
				no_declspec   = False,
				cpp           = True,
				header_name   = "__AKELA_VNA_DLL_HEADER", htype='vna'
			)
	except FileNotFoundError:

		fdir = os.path.dirname(os.path.abspath(__file__))
		loc_header = os.path.join(fdir, "headers", "avmu_header_agg.h")

		if getattr(sys, 'frozen', False):
			# we are running in a |PyInstaller| bundle
			frozenheader = os.path.join(sys._MEIPASS, 'avmu_header_agg.h')
			agg_avmu = open(frozenheader).read()

		elif os.path.exists(loc_header):
			# Allow running from source when outside the development directory structure.
			agg_avmu = open(loc_header).read()
		else:
			raise
	return agg_avmu





from .avmu_exceptions import *
from .avmu_library    import *
from .avmu_utils      import *

##
#  \addtogroup Python-API
#  @{
#
# @brief This is the general import for the entire avmu interface library.
#
#        It is located at the import path of `avmu` (e.g. `import avmu`).
#
#        By including this file, you include (by proxy) the \ref Python-OOP-API
#        and \ref Python-Basic-API contents, which are both wildcard imported (`from x import *`)
#        into this module's namespace.
#
#        Its contents are effectively:
#
#            from .avmu_library    import *
#            from .avmu_exceptions import *
#
#        In general, you should probably not directly import `avmu.avmu_library`, but rather
#        simply `import avmu`, and use it directly.
#
#
#
# @}




import collections
AvmuComboTuple = collections.namedtuple('AvmuComboTuple', ['idx', 'tx_path', 'rx_path'])

def path_the_same(txp, rxp):
	return txp.replace("_TX_", "_RX_") == rxp

def switch_board_type_to_port_list(board_type, is_tx):
	# Switch settings for the TDD and non-TDD boards are the same.
	if board_type == "TDD_4_PORT_SWITCH" or board_type == "SIMPLE_4_PORT_SWITCH":
		if is_tx:
			return ["AVMU_TX_PATH_0", "AVMU_TX_PATH_1", "AVMU_TX_PATH_2", "AVMU_TX_PATH_3"]
		else:
			return ["AVMU_RX_PATH_0", "AVMU_RX_PATH_1", "AVMU_RX_PATH_2", "AVMU_RX_PATH_3"]
	elif board_type == 'NO_SWITCH_BOARD':
		if is_tx:
			return ["AVMU_TX_PATH_0"]
		else:
			return ["AVMU_RX_PATH_1"]
	elif board_type == 'SIMPLE_8_PORT_SWITCH':
		if is_tx:
			return ["AVMU_TX_PATH_0", "AVMU_TX_PATH_1", "AVMU_TX_PATH_2", "AVMU_TX_PATH_3",
					"AVMU_TX_PATH_4", "AVMU_TX_PATH_5", "AVMU_TX_PATH_6", "AVMU_TX_PATH_7"]
		else:
			return ["AVMU_RX_PATH_0", "AVMU_RX_PATH_1", "AVMU_RX_PATH_2", "AVMU_RX_PATH_3",
					"AVMU_RX_PATH_4", "AVMU_RX_PATH_5", "AVMU_RX_PATH_6", "AVMU_RX_PATH_7"]
	elif board_type == 'S_PARAM_SWITCH':
		if is_tx:
			return ["AVMU_TX_PATH_0", "AVMU_TX_PATH_1", "AVMU_TX_PATH_2", "AVMU_TX_PATH_3",
					"AVMU_TX_PATH_4"]
		else:
			return ["AVMU_RX_PATH_0", "AVMU_RX_PATH_1", "AVMU_RX_PATH_2", "AVMU_RX_PATH_3",
					"AVMU_RX_PATH_4"]
	else:
		raise RuntimeError("Invalid switch board type: '%s'" % board_type)

def generate_combo_list(avmu_list, schedule_type):
	combo_listing = []

	# Mask out any disabled avmus so the combos generate properly.
	avmu_list = [tmp for tmp in avmu_list if tmp['AVMU_ENABLE']]

	for tx_avmu in avmu_list:
		tx_avmu_idx = tx_avmu['AVMU_IDX']
		tx_port_list = switch_board_type_to_port_list(tx_avmu['AVMU_SWITCHBOARD_TYPE'], is_tx=True)
		rx_port_list = switch_board_type_to_port_list(tx_avmu['AVMU_SWITCHBOARD_TYPE'], is_tx=False)


		# Since we later require the avmus to have matching
		# switch boards, we can just use the tx avmu for generating both
		for tx_port in tx_port_list:
			for rx_port in rx_port_list:
				tx_set = []

				for rx_avmu in avmu_list:
					rx_avmu_idx = rx_avmu['AVMU_IDX']

					# Sequential scheduling only operates on a per-avmu basis.
					if tx_avmu_idx != rx_avmu_idx and schedule_type == 'SEQUENTIAL':
						continue


					# We need a TDD board to do TDD.
					if tx_avmu['AVMU_DO_TDD'] and tx_avmu['AVMU_SWITCHBOARD_TYPE'] != 'TDD_4_PORT_SWITCH':
						raise RuntimeError("TDD Requires a TDD 4 port switch board! Specified switch board: %s" %
							(tx_avmu['AVMU_SWITCHBOARD_TYPE'], ))

					# Mismatched switch-boards won't do much
					if tx_avmu['AVMU_SWITCHBOARD_TYPE'] != rx_avmu['AVMU_SWITCHBOARD_TYPE']:
						raise RuntimeError("Mismatched switch-board types: %s, %s" %
							(tx_avmu['AVMU_SWITCHBOARD_TYPE'], rx_avmu['AVMU_SWITCHBOARD_TYPE'], ))


					if any([tmp['AVMU_SWITCHBOARD_TYPE'] == 'S_PARAM_SWITCH' for tmp in avmu_list]) and len(avmu_list) > 1:
						raise RuntimeError("S-Param boards can only be run by themselves! Specified switch boards: %s" %
							([tmp['AVMU_SWITCHBOARD_TYPE'] for tmp in avmu_list], ))

					if tx_avmu['AVMU_SWITCHBOARD_TYPE'] == 'S_PARAM_SWITCH':
						tx_num = tx_port.split("_")[-1]
						rx_num = rx_port.split("_")[-1]
						# The valid combos for the s-param board are 0->0, 1->1, 2->2, 3->3, 4->4.

						if tx_num == rx_num:
							avmu_conf = AvmuComboTuple(rx_avmu_idx, tx_port, rx_port)
							tx_set.append(avmu_conf)

					else:
						# Normal switch boards
						is_transmit = tx_avmu_idx == rx_avmu_idx

						# We can't do monostatic measurements without TDD
						if tx_avmu_idx == rx_avmu_idx and path_the_same(tx_port, rx_port):
							rx_port_masked = "AVMU_RX_PATH_NONE"
						else:
							rx_port_masked = rx_port

						if not is_transmit:
							tx_port_masked = "AVMU_TX_PATH_NONE"
						else:
							tx_port_masked = tx_port

						avmu_conf = AvmuComboTuple(rx_avmu_idx, tx_port_masked, rx_port_masked)
						if schedule_type == 'SIMULTANEOUS':
							tx_set.append(avmu_conf)
						elif schedule_type == "SEQUENTIAL":
							# Since sequential scheduling is always 1 tx, 1 rx,
							# the combo's where the rx path is RX_PATH_NONE wind
							# up being useless. As such, skip them.
							if rx_port_masked != "AVMU_RX_PATH_NONE":
								tx_set.append(avmu_conf)
						else:
							raise ValueError("Unknown Scheduling mode: %s" % (schedule_type, ))


				if not tx_set:
					# If no paths were generated for the combo,
					# don't append it to the output (this happens
					# mostly in the sequential mode).
					pass
				# If we're doing single-avmu measurements, we don't need to bother
				# with the rx_none paths, since there's no other avmu doing RX.
				elif len(avmu_list) == 1 and tx_set[0].rx_path == 'AVMU_RX_PATH_NONE':
					pass
				else:
					# print("Adding: ", tx_set)
					combo_listing.append(tuple(tx_set))

	for combo in combo_listing:
		fail = any([tmp.rx_path != "AVMU_RX_PATH_NONE" for tmp in combo])
		# if not fail:
		# 	print("Passed conf: ")
		# 	print(avmu_list)
		# 	print(schedule_type)
		# 	print("Computed combos:")
		# 	for it_combo in combo_listing:
		# 		print("	", it_combo)
		assert fail, "At least one avmu must be receiving in every combo: %s" % combo
	return combo_listing





import sys
import struct
import os

if sys.version_info < (3, 4):
    sys.exit('Sorry, Python < 3.4 is not supported')

if os.name != 'nt':
    sys.exit('Only windows installs are supported for PyPi installs. Please ' +
        'contact AKELA Inc for linux packages.')

if struct.calcsize("P") * 8 != 64:
    sys.exit("This interface requires 64 bit python!")


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name                          = "avmu",
    version                       = "0.0.6",
    author                        = "Connor Wolf, Akela Inc",
    author_email                  = "cwolf@akelainc.com",
    description                   = "Control interface and API for running Akela Vector Measurement Units.",
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    url                           = "https://github.com/AkelaInc/avmu",
    packages                      = setuptools.find_packages(),
    python_requires               = ">=3.4",
    install_requires              = [
            'numpy',
            'cffi',
        ],
    include_package_data          = True,
    classifiers                   = [
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
    ],
)




'''

Each exception corresponds to the eponymous C error return code, prefixed with ``Avmu_Exception_``.

Note that these exceptions are also available as ``avmu.Avmu_Exception_*``, as they're
star-imported into the main avmu module.

'''


class Avmu_Exception(Exception):
	'''
	Base exception class that all library exceptions inherit from. This
	can be used to easily catch all exceptions that are specifically
	thrown by the ``avmu`` library.
	'''
	pass


class Avmu_Exception_Bad_Atten(Avmu_Exception):
	'''
	Error code indicating the specified attenuation value was not valid.
	'''
	pass
class Avmu_Exception_Bad_Cal(Avmu_Exception):
	'''
	Error code indicating the current calibration is invalid. Only relevant for VNAs.
	'''
	pass
class Avmu_Exception_Bad_Handle(Avmu_Exception):
	'''
	Error code indicating the current handle passed to a DLL function was not valid.

	Generally indicates an internal error. If this occurs repeatedly, please contact support.
	'''
	pass
class Avmu_Exception_Bad_Hop(Avmu_Exception):
	'''
	Error code indicating the specified hop-rate value was not valid.
	'''
	pass
class Avmu_Exception_Bad_Path(Avmu_Exception):
	'''
	Error code indicating the specified path is not allowed, or was invalid.
	'''
	pass
class Avmu_Exception_Bad_Prom(Avmu_Exception):
	'''
	Error code indicating that the AVMU PROM was not in a known format.
	Please contact support if this is a reoccurring issue.
	'''
	pass
class Avmu_Exception_Bytes(Avmu_Exception):
	'''
	Error code indicating the DLL has received an incorrect number of bytes from the AVMU.
	Possibly indicative of a network issue.
	'''
	pass
class Avmu_Exception_Freq_Out_Of_Bounds(Avmu_Exception):
	'''
	Error code indicating one of the specified frequency bounds was beyond the hardware's available range.
	'''
	pass
class Avmu_Exception_Interrupted(Avmu_Exception):
	'''
	Error code indicating that the measurement was asynchronously interrupted from another thread.
	'''
	pass
class Avmu_Exception_No_Response(Avmu_Exception):
	'''
	Error code indicating the remote instrument failed to respond to commands within the current timeout period.
	If this persists, check your network and try power cycling the AVMU.
	'''
	pass
class Avmu_Exception_Missing_Ip(Avmu_Exception):
	'''
	Error code indicating you have not yet specified the IP of the remote device before connecting.
	'''
	pass
class Avmu_Exception_Missing_Port(Avmu_Exception):
	'''
	Error code indicating you have not yet specified the port for the remote device before connecting.
	'''
	pass
class Avmu_Exception_Missing_Hop(Avmu_Exception):
	'''
	Error code indicating you have not yet specified the hop-rate for the remote device before connecting.
	'''
	pass
class Avmu_Exception_Missing_Atten(Avmu_Exception):
	'''
	The switchboard config requires an attenuation value to be specified, and it has not been.
	'''
	pass
class Avmu_Exception_Missing_Freqs(Avmu_Exception):
	'''
	Error code indicating you have not yet specified the frequency endpoints for the remote device before connecting.
	'''
	pass
class Avmu_Exception_Prog_Overflow(Avmu_Exception):
	'''
	Error code indicating the generated program to run on the remote device was larger than it can handle.
	'''
	pass
class Avmu_Exception_Socket(Avmu_Exception):
	'''
	Error code indicating an unknown socket error occured.
	'''
	pass
class Avmu_Exception_Too_Many_Points(Avmu_Exception):
	'''
	Error code indicating you are trying to take more data points than the hardware is capable of.
	'''
	pass
class Avmu_Exception_Wrong_State(Avmu_Exception):
	'''
	Error code indicating the hardware is not in the correct state for the function you are trying to call.
	'''
	pass
class Avmu_Exception_Empty_Prom(Avmu_Exception):
	'''
	Error code indicating the remote device's PROM appears to be empty.
	Power cycle the AVMU, and if the problem persists, contact Akela.
	'''
	pass
class Avmu_Exception_Path_Already_Measured(Avmu_Exception):
	'''
	Error code indicating you are adding a path to measure that you had already added.

	These restrictions have been relaxed in recent releases.
	'''
	pass
class Avmu_Exception_No_Measured_Paths(Avmu_Exception):
	'''
	Error code indicating there are no paths to measure, and the hardware has nothing to do.
	'''
	pass
class Avmu_Exception_Wrong_Program_Type(Avmu_Exception):
	'''
	Error code indicating you are in sync mode and are trying to call an async startup function.
	'''
	pass
class Avmu_Exception_Unknown_Feature(Avmu_Exception):
	'''
	Error code indicating the AVMU's prom indicates it has a hardware feature that this
	version of the DLL doesn't know about. Either indicates you need to update, or the prom is corrupted.
	'''
	pass
class Avmu_Exception_Feature_Not_Present(Avmu_Exception):
	'''
	You are trying to configure a hardware feature not present in the connected AVMU.
	'''
	pass
class Avmu_Exception_No_Attenuator_Present(Avmu_Exception):
	'''
	You are trying to specify an attenuation value on hardware without an attenuator.
	'''
	pass
class Avmu_Exception_Bad_IP_Port(Avmu_Exception):
	'''
	The UDP port you specified is not valid. You must use a port >= 1024 and < 65535
	'''
	pass
class Avmu_Exception_Task_Array_Invalid(Avmu_Exception):
	'''
	The batch-task array parameter appears to not be valid.
	'''
	pass
class Avmu_Exception_Path_Has_No_Data(Avmu_Exception):
	'''
	No data was acquired for the specified path. Either you didn't add the path,
	didn't call measure(), or something else is wrong.
	'''
	pass
class Avmu_Exception_Err_Index_Out_Of_Bounds(Avmu_Exception):
	'''
	 The sweep index/exclusion band you are trying to extract is outside the bounds of the array.
	'''
	pass
class Avmu_Exception_Err_Invalid_Parameter(Avmu_Exception):
	'''
	 One of the specified parameters is not valid.
	 If you are specifying sweep frequency, check your frequency bands.
	'''
	pass
class Avmu_Exception_Err_Prom_Invalid_Feature_Configuration(Avmu_Exception):
	'''
	 The AVMU's prom contains feature flags that are incompatible with each other.
	  Please contact support.
	'''
	pass



## @}



'''
# #########################################################################

Python api to AKELA Inc's AVMU hardware.

Note: This library is a relatively thin wrapper on top of
the C/C++ AVMU DLL interface. In general, the documentation
for the C interface is more complete, and generally more likely
to be kept completely up to date. Please check that documentation
in case of ambiguities, before contacting support.

Author: Connor Wolf <cwolf@akelainc.com>

# #########################################################################
'''

import logging
import numpy as np
from . import dll_loader
from . import avmu_exceptions

class AvmuInterface(object):



	def __init__(self, share_from_interface = None):
		'''
		Create the base AVMU interface class.

		This is the core class for acquiring data via an AKELA Inc. Vector
		Measurement Unit (or AVMU).

		For multi-AVMU operation, pass the first created interface
		to any additional interface class instances. This allows all
		the interface instances to share the same underlying communication
		objects, which is needed for synchronized operation.

		Note: If you don't share the communcation object, calls
		such as broadcastBeginCommand() will not operate properly.

		Unfortunately, there is no coherent way to validate this
		in the library code, so it's left up to the user.

		Accessing the same AVMU interface from separate threads is *not*
		threadsafe. However, Accessing different AVMU interface classes from
		separate threads that share the same underlying communication interface
		should be thread-safe.

		Note: All library functions can theoretically raise  :class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Handle`
		if the internal task handle has been corrupted. As such, this particular exception will not be explicitly
		enumerated for every possible call.

		'''


		self.log = logging.getLogger("Main.Dll")

		self.log.debug("Loading DLL")

		self.ffi, self.dll = dll_loader.load_ffi_interface()

		if "Private API mode" in self.versionString():
			from . import private_api
			self.__class__ = private_api.PromOverrideAvmuInterface

		self.log.debug("Constructing constant mapping tables.")
		self.___construct_map_tables()

		self.task_handle = self.__createTask(share_from_interface)

		self.measured_paths = []

		self.serial_buf_sz = 0
		self.active_receivers = [0]

	def __del__(self):
		try:
			self.__deleteTask(self.task_handle)
		except AttributeError:
			print("WARNING: Error when trying to delete task handle!")

	def __repr__(self):
		ret = "<{} for radar {}:{} state {}, handle {}>".format(
			self.__class__.__name__,
			self.getIPAddress(),
			self.getIPPort(),
			self.getState(),

			# This is legit kind of horrible.
			# Only thing that makes it even marginally not-bad is that
			# it's only for printing an error message.
			str(self.task_handle).strip(">").split(" ")[-1]
			)

		return ret


	def __getRawTaskHandle(self):
		self.log.debug("__getRawTaskHandle call")
		return self.task_handle

	def __createTask(self, fromtask):
		self.log.debug("__createTask call")
		# Signature: TaskHandle createTask();
		self.log.debug("Creating task.")
		if fromtask:
			th = self.dll.createSharedTask(fromtask.__getRawTaskHandle())
		else:
			th = self.dll.createTask()

		return th

	def __deleteTask(self, task_handle):
		self.log.debug("__deleteTask call")
		# Signature: void deleteTask(TaskHandle t);
		self.log.info("Destroying task.")
		self.dll.deleteTask(task_handle)


	def ___construct_map_tables(self):




		self.log.debug("___construct_map_tables call")
		self.errors = {
			self.dll.ERR_OK                                 : None,
			self.dll.ERR_BAD_ATTEN                          : avmu_exceptions.Avmu_Exception_Bad_Atten,
			self.dll.ERR_BAD_CAL                            : avmu_exceptions.Avmu_Exception_Bad_Cal,
			self.dll.ERR_BAD_HANDLE                         : avmu_exceptions.Avmu_Exception_Bad_Handle,
			self.dll.ERR_BAD_HOP                            : avmu_exceptions.Avmu_Exception_Bad_Hop,
			self.dll.ERR_BAD_PATH                           : avmu_exceptions.Avmu_Exception_Bad_Path,
			self.dll.ERR_BAD_PROM                           : avmu_exceptions.Avmu_Exception_Bad_Prom,
			self.dll.ERR_BYTES                              : avmu_exceptions.Avmu_Exception_Bytes,
			self.dll.ERR_EMPTY_PROM                         : avmu_exceptions.Avmu_Exception_Empty_Prom,
			self.dll.ERR_FEATURE_NOT_PRESENT                : avmu_exceptions.Avmu_Exception_Feature_Not_Present,
			self.dll.ERR_FREQ_OUT_OF_BOUNDS                 : avmu_exceptions.Avmu_Exception_Freq_Out_Of_Bounds,
			self.dll.ERR_INTERRUPTED                        : avmu_exceptions.Avmu_Exception_Interrupted,
			self.dll.ERR_MISSING_ATTEN                      : avmu_exceptions.Avmu_Exception_Missing_Atten,
			self.dll.ERR_MISSING_FREQS                      : avmu_exceptions.Avmu_Exception_Missing_Freqs,
			self.dll.ERR_MISSING_HOP                        : avmu_exceptions.Avmu_Exception_Missing_Hop,
			self.dll.ERR_MISSING_IP                         : avmu_exceptions.Avmu_Exception_Missing_Ip,
			self.dll.ERR_MISSING_PORT                       : avmu_exceptions.Avmu_Exception_Missing_Port,
			self.dll.ERR_NO_ATTEN_PRESENT                   : avmu_exceptions.Avmu_Exception_No_Attenuator_Present,
			self.dll.ERR_NO_PATHS_MEASURED                  : avmu_exceptions.Avmu_Exception_No_Measured_Paths,
			self.dll.ERR_NO_RESPONSE                        : avmu_exceptions.Avmu_Exception_No_Response,
			self.dll.ERR_PATH_ALREADY_MEASURED              : avmu_exceptions.Avmu_Exception_Path_Already_Measured,
			self.dll.ERR_PROG_OVERFLOW                      : avmu_exceptions.Avmu_Exception_Prog_Overflow,
			self.dll.ERR_SOCKET                             : avmu_exceptions.Avmu_Exception_Socket,
			self.dll.ERR_TOO_MANY_POINTS                    : avmu_exceptions.Avmu_Exception_Too_Many_Points,
			self.dll.ERR_UNKNOWN_FEATURE                    : avmu_exceptions.Avmu_Exception_Unknown_Feature,
			self.dll.ERR_WRONG_PROGRAM_TYPE                 : avmu_exceptions.Avmu_Exception_Wrong_Program_Type,
			self.dll.ERR_WRONG_STATE                        : avmu_exceptions.Avmu_Exception_Wrong_State,
			self.dll.ERR_BAD_IP_PORT                        : avmu_exceptions.Avmu_Exception_Bad_IP_Port,
			self.dll.ERR_TASK_ARRAY_INVALID                 : avmu_exceptions.Avmu_Exception_Task_Array_Invalid,
			self.dll.ERR_PATH_HAS_NO_DATA                   : avmu_exceptions.Avmu_Exception_Path_Has_No_Data,
			# self.dll.ERR_INDEX_OUT_OF_BOUNDS                : avmu_exceptions.Avmu_Exception_Err_Index_Out_Of_Bounds,
			# self.dll.ERR_INVALID_PARAMETER                  : avmu_exceptions.Avmu_Exception_Err_Invalid_Parameter,
			# self.dll.ERR_PROM_INVALID_FEATURE_CONFIGURATION : avmu_exceptions.Avmu_Exception_Err_Prom_Invalid_Feature_Configuration,

		}

		self.hops = {
			'HOP_UNDEFINED' : self.dll.HOP_UNDEFINED,
			'HOP_90K'       : self.dll.HOP_90K,
			'HOP_45K'       : self.dll.HOP_45K,
			'HOP_30K'       : self.dll.HOP_30K,
			'HOP_15K'       : self.dll.HOP_15K,
			'HOP_7K'        : self.dll.HOP_7K,
			'HOP_3K'        : self.dll.HOP_3K,
			'HOP_2K'        : self.dll.HOP_2K,
			'HOP_1K'        : self.dll.HOP_1K,
			'HOP_550'       : self.dll.HOP_550,
			'HOP_312'       : self.dll.HOP_312,
			'HOP_156'       : self.dll.HOP_156,
			'HOP_78'        : self.dll.HOP_78,
			'HOP_39'        : self.dll.HOP_39,
			'HOP_20'        : self.dll.HOP_20,
		}


		self.tx_paths = {
			'AVMU_TX_PATH_0'    : self.dll.AVMU_TX_PATH_0,
			'AVMU_TX_PATH_1'    : self.dll.AVMU_TX_PATH_1,
			'AVMU_TX_PATH_2'    : self.dll.AVMU_TX_PATH_2,
			'AVMU_TX_PATH_3'    : self.dll.AVMU_TX_PATH_3,
			'AVMU_TX_PATH_4'    : self.dll.AVMU_TX_PATH_4,
			'AVMU_TX_PATH_5'    : self.dll.AVMU_TX_PATH_5,
			'AVMU_TX_PATH_6'    : self.dll.AVMU_TX_PATH_6,
			'AVMU_TX_PATH_7'    : self.dll.AVMU_TX_PATH_7,
			'AVMU_TX_PATH_NONE' : self.dll.AVMU_TX_PATH_NONE,
		}
		self.rx_paths = {
			'AVMU_RX_PATH_0'    : self.dll.AVMU_RX_PATH_0,
			'AVMU_RX_PATH_1'    : self.dll.AVMU_RX_PATH_1,
			'AVMU_RX_PATH_2'    : self.dll.AVMU_RX_PATH_2,
			'AVMU_RX_PATH_3'    : self.dll.AVMU_RX_PATH_3,
			'AVMU_RX_PATH_4'    : self.dll.AVMU_RX_PATH_4,
			'AVMU_RX_PATH_5'    : self.dll.AVMU_RX_PATH_5,
			'AVMU_RX_PATH_6'    : self.dll.AVMU_RX_PATH_6,
			'AVMU_RX_PATH_7'    : self.dll.AVMU_RX_PATH_7,
			'AVMU_RX_PATH_NONE' : self.dll.AVMU_RX_PATH_NONE,
		}


		self.tx_paths_int = {
			'AVMU_TX_PATH_0'    : 0,
			'AVMU_TX_PATH_1'    : 1,
			'AVMU_TX_PATH_2'    : 2,
			'AVMU_TX_PATH_3'    : 3,
			'AVMU_TX_PATH_4'    : 4,
			'AVMU_TX_PATH_5'    : 5,
			'AVMU_TX_PATH_6'    : 6,
			'AVMU_TX_PATH_7'    : 7,
			'AVMU_TX_PATH_NONE' : -1,
		}
		self.rx_paths_int = {
			'AVMU_RX_PATH_0'    :  0,
			'AVMU_RX_PATH_1'    :  1,
			'AVMU_RX_PATH_2'    :  2,
			'AVMU_RX_PATH_3'    :  3,
			'AVMU_RX_PATH_4'    :  4,
			'AVMU_RX_PATH_5'    :  5,
			'AVMU_RX_PATH_6'    :  6,
			'AVMU_RX_PATH_7'    :  7,
			'AVMU_RX_PATH_NONE' : -1,
		}

		self.tx_paths_int_enum = {
			 0 : self.dll.AVMU_TX_PATH_0,
			 1 : self.dll.AVMU_TX_PATH_1,
			 2 : self.dll.AVMU_TX_PATH_2,
			 3 : self.dll.AVMU_TX_PATH_3,
			 4 : self.dll.AVMU_TX_PATH_4,
			 5 : self.dll.AVMU_TX_PATH_5,
			 6 : self.dll.AVMU_TX_PATH_6,
			 7 : self.dll.AVMU_TX_PATH_7,
			-1 : self.dll.AVMU_TX_PATH_NONE,
		}

		self.rx_paths_int_enum = {
			 0 : self.dll.AVMU_RX_PATH_0,
			 1 : self.dll.AVMU_RX_PATH_1,
			 2 : self.dll.AVMU_RX_PATH_2,
			 3 : self.dll.AVMU_RX_PATH_3,
			 4 : self.dll.AVMU_RX_PATH_4,
			 5 : self.dll.AVMU_RX_PATH_5,
			 6 : self.dll.AVMU_RX_PATH_6,
			 7 : self.dll.AVMU_RX_PATH_7,
			-1 : self.dll.AVMU_RX_PATH_NONE,
		}

		self.tx_paths_enum_int = {
			self.dll.AVMU_TX_PATH_0    :  0,
			self.dll.AVMU_TX_PATH_1    :  1,
			self.dll.AVMU_TX_PATH_2    :  2,
			self.dll.AVMU_TX_PATH_3    :  3,
			self.dll.AVMU_TX_PATH_4    :  4,
			self.dll.AVMU_TX_PATH_5    :  5,
			self.dll.AVMU_TX_PATH_6    :  6,
			self.dll.AVMU_TX_PATH_7    :  7,
			self.dll.AVMU_TX_PATH_NONE : -1,

		}

		self.rx_paths_enum_int = {
			self.dll.AVMU_RX_PATH_0    :  0,
			self.dll.AVMU_RX_PATH_1    :  1,
			self.dll.AVMU_RX_PATH_2    :  2,
			self.dll.AVMU_RX_PATH_3    :  3,
			self.dll.AVMU_RX_PATH_4    :  4,
			self.dll.AVMU_RX_PATH_5    :  5,
			self.dll.AVMU_RX_PATH_6    :  6,
			self.dll.AVMU_RX_PATH_7    :  7,
			self.dll.AVMU_RX_PATH_NONE : -1,
		}

		self.tx_paths_enum_str = {
			self.dll.AVMU_TX_PATH_0    : 'AVMU_TX_PATH_0',
			self.dll.AVMU_TX_PATH_1    : 'AVMU_TX_PATH_1',
			self.dll.AVMU_TX_PATH_2    : 'AVMU_TX_PATH_2',
			self.dll.AVMU_TX_PATH_3    : 'AVMU_TX_PATH_3',
			self.dll.AVMU_TX_PATH_4    : 'AVMU_TX_PATH_4',
			self.dll.AVMU_TX_PATH_5    : 'AVMU_TX_PATH_5',
			self.dll.AVMU_TX_PATH_6    : 'AVMU_TX_PATH_6',
			self.dll.AVMU_TX_PATH_7    : 'AVMU_TX_PATH_7',
			self.dll.AVMU_TX_PATH_NONE : 'AVMU_TX_PATH_NONE',

		}

		self.rx_paths_enum_str = {
			self.dll.AVMU_RX_PATH_0    : 'AVMU_RX_PATH_0',
			self.dll.AVMU_RX_PATH_1    : 'AVMU_RX_PATH_1',
			self.dll.AVMU_RX_PATH_2    : 'AVMU_RX_PATH_2',
			self.dll.AVMU_RX_PATH_3    : 'AVMU_RX_PATH_3',
			self.dll.AVMU_RX_PATH_4    : 'AVMU_RX_PATH_4',
			self.dll.AVMU_RX_PATH_5    : 'AVMU_RX_PATH_5',
			self.dll.AVMU_RX_PATH_6    : 'AVMU_RX_PATH_6',
			self.dll.AVMU_RX_PATH_7    : 'AVMU_RX_PATH_7',
			self.dll.AVMU_RX_PATH_NONE : 'AVMU_RX_PATH_NONE',
		}


		self.prog_type = {
			'PROG_ASYNC'          : self.dll.PROG_ASYNC,
			'PROG_SYNC'           : self.dll.PROG_SYNC,

		}
		self.run_state = {
			'TASK_RUNNING'        : self.dll.TASK_RUNNING,
			'TASK_STARTED'        : self.dll.TASK_STARTED,
			'TASK_STOPPED'        : self.dll.TASK_STOPPED,
			'TASK_UNINITIALIZED'  : self.dll.TASK_UNINITIALIZED,
		}
		self.sync_pulse_mode = {
			'SYNC_IGNORE'         : self.dll.SYNC_IGNORE,
			'SYNC_GENERATE'       : self.dll.SYNC_GENERATE,
			'SYNC_RECEIVE'        : self.dll.SYNC_RECEIVE,
		}

		self.if_gain_settings = {
				'AVMU_GAIN_USE_DEFAULT' : self.dll.AVMU_GAIN_USE_DEFAULT,
				'AVMU_GAIN_0'           : self.dll.AVMU_GAIN_0,
				'AVMU_GAIN_3'           : self.dll.AVMU_GAIN_3,
				'AVMU_GAIN_6'           : self.dll.AVMU_GAIN_6,
				'AVMU_GAIN_9'           : self.dll.AVMU_GAIN_9,
				'AVMU_GAIN_12'          : self.dll.AVMU_GAIN_12,
				'AVMU_GAIN_15'          : self.dll.AVMU_GAIN_15,
				'AVMU_GAIN_18'          : self.dll.AVMU_GAIN_18,
				'AVMU_GAIN_21'          : self.dll.AVMU_GAIN_21,
				'AVMU_GAIN_24'          : self.dll.AVMU_GAIN_24,
				'AVMU_GAIN_27'          : self.dll.AVMU_GAIN_27,
				'AVMU_GAIN_30'          : self.dll.AVMU_GAIN_30,
				'AVMU_GAIN_33'          : self.dll.AVMU_GAIN_33,
				'AVMU_GAIN_36'          : self.dll.AVMU_GAIN_36,
				'AVMU_GAIN_39'          : self.dll.AVMU_GAIN_39,
				'AVMU_GAIN_42'          : self.dll.AVMU_GAIN_42,
				'AVMU_GAIN_45'          : self.dll.AVMU_GAIN_45,
		}

		self.if_gain_inverse = {
				self.dll.AVMU_GAIN_USE_DEFAULT : 'AVMU_GAIN_USE_DEFAULT',
				self.dll.AVMU_GAIN_0           : 'AVMU_GAIN_0',
				self.dll.AVMU_GAIN_3           : 'AVMU_GAIN_3',
				self.dll.AVMU_GAIN_6           : 'AVMU_GAIN_6',
				self.dll.AVMU_GAIN_9           : 'AVMU_GAIN_9',
				self.dll.AVMU_GAIN_12          : 'AVMU_GAIN_12',
				self.dll.AVMU_GAIN_15          : 'AVMU_GAIN_15',
				self.dll.AVMU_GAIN_18          : 'AVMU_GAIN_18',
				self.dll.AVMU_GAIN_21          : 'AVMU_GAIN_21',
				self.dll.AVMU_GAIN_24          : 'AVMU_GAIN_24',
				self.dll.AVMU_GAIN_27          : 'AVMU_GAIN_27',
				self.dll.AVMU_GAIN_30          : 'AVMU_GAIN_30',
				self.dll.AVMU_GAIN_33          : 'AVMU_GAIN_33',
				self.dll.AVMU_GAIN_36          : 'AVMU_GAIN_36',
				self.dll.AVMU_GAIN_39          : 'AVMU_GAIN_39',
				self.dll.AVMU_GAIN_42          : 'AVMU_GAIN_42',
				self.dll.AVMU_GAIN_45          : 'AVMU_GAIN_45',
		}


	def __check_ret(self, ret_val):
		self.log.debug("__check_ret call")
		try:
			state = self.getState()
		except Exception:
			self.log.error("Failed to query current state!")
			state = "Unknown!"

		assert ret_val in self.errors, "Unknown returned error code: %s. Current state: %s" % (ret_val, state)
		err = self.errors[ret_val]
		if err:
			raise err("Call returned error value: %s. Current state: %s" % (ret_val, state))


	#################################################################################
	#        Configuration
	#################################################################################

	def addPathToMeasure(self, tx_path, rx_path, who_is_transmitting=None, port_is_transmitting=None):
		'''
		Add a path to measure.

		Adds a path to the list of active paths that will be
		measured when data is actually acquired.

		Args:
			tx_path (str): Transmitting port of the path to add. Note that
			                 this can very well be TX_NONE, if your application is
			                 using multiple AVMUs, and one of the other AVMUs is
			                 the one actively transmitting.
			rx_path (str): RX Path for the combo. This can be RX_NONE if there
			                is no need for the associated rx data. Note that RX_NONE
			                will still return data (as the transmitter will still
			                need to do the associated frequency stepping), but the
			                return content will be only cross-board leakage within
			                the AVMU.

		Returns:
			Nothing

		raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Path`:   if a path value specified is invalid.

		'''
		self.log.debug("addPathToMeasure call")
		# Signature: ErrCode addPathToMeasure(TaskHandle t, RFPath path);
		if not tx_path in self.tx_paths: raise avmu_exceptions.Avmu_Exception_Bad_Path("Invalid TX Path: '%s'" % (tx_path, ))
		if not rx_path in self.rx_paths: raise avmu_exceptions.Avmu_Exception_Bad_Path("Invalid RX Path: '%s'" % (rx_path, ))

		items = [
			("tx_path",              tx_path, str),
			("rx_path",              rx_path, str),
			("who_is_transmitting",  who_is_transmitting, (int, type(None))),
			("port_is_transmitting", port_is_transmitting, (int, type(None)))
		]

		for name, value, expected_type in items:
			assert isinstance(value, expected_type), "Parameter %s is not of expected type (%s). Passed type %s, value %s" % (
				name, expected_type, type(value), value)

		self.log.debug("Adding path to measure: %s -> %s (transmiting:  %s -> %s)", tx_path, rx_path, who_is_transmitting, port_is_transmitting)
		self.measured_paths.append((who_is_transmitting, port_is_transmitting, self.tx_paths_int[tx_path], self.rx_paths_int[rx_path]))
		ret = self.dll.addPathToMeasure(self.task_handle, self.tx_paths[tx_path], self.rx_paths[rx_path])
		self.__check_ret(ret)

	def clearMeasuredPaths(self):
		'''
		Clear list of paths being measured.

		Returns:
			Nothing

		'''
		self.log.debug("clearMeasuredPaths call")
		# Signature: ErrCode clearMeasuredPaths(TaskHandle t);
		self.log.debug("Clearing measured paths")
		self.measured_paths = []
		ret = self.dll.clearMeasuredPaths(self.task_handle)
		self.__check_ret(ret)

	###############################################################

	def getFrequencies(self):
		'''

		Get a list containing the actual frequencies the hardware will sample for
		the configured sweep in task `t`. Units are MHz.

		The actual frequency points can differ from the requested frequency points
		because the hardware has fixed precision, and cannot achieve every arbitrary
		frequency value within its tunable bands. The values in this list are the
		requested frequency points after snapping them to the closest achievable
		frequency.

		Returns:
			list of floating point frequencies, in MHz


		'''
		self.log.debug("getFrequencies call")
		# Signature: ErrCode getFrequencies(TaskHandle t, double* freqs, int pts_in_freqs);
		npts = self.getNumberOfFrequencies()
		freq_arr = self.ffi.new("double[] ", [0] * npts)
		ret = self.dll.getFrequencies(self.task_handle, freq_arr, npts)
		self.__check_ret(ret)
		return list(freq_arr)

	def setFrequencies(self, freqs):
		'''
		Set the frequencies to measure during each sweep. Units are MHz.

		Note that the AVMU frequency generation
		hardware has fixed precision and so the generated frequency may not be exactly
		equal to the requested frequency. This function silently converts all requested
		frequencies to frequencies that can be exactly generated by the hardware.
		This has important implications for Doppler noise when doing a linear sweep.
		AKELA recommends using the function :func:`utilFixLinearSweepLimits()` to ensure
		every frequency is exactly generateable and that the frequencies are equally
		spaced. Use the :func:`getFrequencies()` function to get the actual frequencies being
		generated.

		Args:
			freqs (list): array of frequencies to sample, in MHz

		Returns:
			Nothing

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Too_Many_Points` if N is larger than \
			                                the maximum allowed (see HardwareDetails)
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Freq_Out_Of_Bounds` if a frequency is \
			                                beyond the allowed min/max. (You can get the min and max from \
			                                the HardwareDetails struct returned by :func:`getHardwareDetails()`)
		'''
		self.log.debug("setFrequencies call")
		# Signature: ErrCode setFrequencies(TaskHandle t, const double* freqs, const unsigned int N);
		freq_arr = self.ffi.new("double[] ", freqs)
		ret = self.dll.setFrequencies(self.task_handle, freq_arr, len(freqs))
		self.__check_ret(ret)

	###############################################################

	def getHopRate(self):
		'''
		Get the frequency hopping rate associated with this Task object.
		If no rate has yet been set, this function returns ``HOP_UNDEFINED``.

		Returns:
			Current hop rate as a string.
		'''

		self.log.debug("getHopRate call")
		# Signature: HopRate getHopRate(TaskHandle t);
		ret = self.dll.getHopRate(self.task_handle)
		for key, value in self.hops.items():
			if value == ret:
				self.log.debug("Current hop rate: %s", key)
				return key
		raise avmu_exceptions.Avmu_Exception_Missing_Hop("getHopRate() returned an unknown hop-rate value: %s" % ret)

	def setHopRate(self, hop_str):
		'''
		Set the frequency hopping rate. See the values defined in the "HopRate" type above.

		Args:
			hop_str (str): The new hop rate.

		Returns:
			Nothing

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Hop` if there was something wrong with the hop rate parameter
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the ``TASK_UNINITIALIZED`` or ``TASK_STOPPED`` state
		'''

		self.log.debug("setHopRate call")
		# Signature: ErrCode setHopRate(TaskHandle t, const HopRate rate);
		self.log.debug("Setting hop rate to: %s", hop_str)
		assert hop_str != "HOP_UNDEFINED", "You cannot set the hop rate to undefined!"
		assert hop_str in self.hops, "Invalid hop rate: '%s'!" % hop_str
		ret = self.dll.setHopRate(self.task_handle, self.hops[hop_str])
		self.__check_ret(ret)

	###############################################################

	def getIPAddress(self):
		'''
		Get the current AVMU IP address for the Task object.
		When no IP has yet been set, this will return a NULL
		char*.

		Returns:
			string of the current remote IP. If the IP is unset, returns ``None``
		'''

		self.log.debug("getIPAddress call")
		# Signature: const char* getIPAddress(TaskHandle t);
		ret = self.dll.getIPAddress(self.task_handle)

		if ret == self.ffi.NULL:
			self.log.debug("Null remote IP!")
			return None
		ret = self.ffi.string(ret).decode("ascii")
		self.log.debug("Current remote IP address: %s", ret)
		return ret

	def setIPAddress(self, ip_address):
		'''

		Sets the IPv4 address of the unit to communicate with.  \
		The ipv4 parameter is copied into the Task's memory.  \
		On success the Task's state will be ``TASK_UNINITIALIZED``. \

		Args:
			ip_address (str) IP representation, e.g. "192.168.1.207", etc...

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State`if the Task is not in the ``TASK_UNINITIALIZED`` or ``TASK_STOPPED`` state

		'''
		self.log.debug("setIPAddress call")
		# Signature: ErrCode setIPAddress(TaskHandle t, const char* ipv4);
		assert ip_address.strip("0123456789.") == "", "Invalid characters in IP: '%s' (full string: '%s')" % (ip_address.strip("0123456789."), ip_address)
		self.log.debug("Setting remote IP address to: %s", ip_address)
		ret = self.dll.setIPAddress(self.task_handle, ip_address.encode("ascii"))
		self.__check_ret(ret)

	###############################################################

	def getIPPort(self):
		'''
		Get the current port for IP communications. When uninitialized, this will default to 0.
		'''
		self.log.debug("getIPPort call")
		# Signature: int getIPPort(TaskHandle t);
		port = self.dll.getIPPort(self.task_handle)
		self.log.debug("Current remote IP port: %s", port)
		return port

	def setIPPort(self, port):
		'''
		 Sets the port on which to communicate with the unit.

		The value of `port` MUST be > 1024, as 1024 is reserved
		for broadcast operations.

		Note that for multi-unit configurations, each unit must be
		assigned a unique port > 1024 (the AVMU units listen on all
		ports, but only respond to broadcast commands on port 1024).
		On success, the Task's state will be ``TASK_UNINITIALIZED``.

		Args:
			port (int) port number

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State`if the Task is not in the ``TASK_UNINITIALIZED`` or ``TASK_STOPPED`` state
		'''

		self.log.debug("setIPPort call")
		# Signature: ErrCode setIPPort(TaskHandle t, const int port);
		self.log.debug("Setting remote IP port to: %s", port)
		ret = self.dll.setIPPort(self.task_handle, port)
		self.__check_ret(ret)

	###############################################################

	def getMeasurementType(self):
		'''
		Get the current measurement ProgramType type.

		Returns:
			(str) Current measurement type.
		'''
		self.log.debug("getMeasurementType call")
		# Signature: ProgramType getMeasurementType(TaskHandle t);
		prog = self.dll.getMeasurementType(self.task_handle)

		for prog_name, prog_val in self.prog_type.items():
			if prog_val == prog:
				return prog_name

		raise avmu_exceptions.Avmu_Exception_Wrong_Program_Type("Unknown program type value (%s)!" % prog)

	def setMeasurementType(self, measure_type):
		'''
		Set the measurement type (synchronous or asynchronous). Note that the AVMU
		defaults to synchonous (PROG_SYNC).

		Args:
			measure_type (str) One of the ProgramType types (``PROG_ASYNC`` or ``PROG_SYNC``).
		'''
		self.log.debug("setMeasurementType call")
		# Signature: ErrCode setMeasurementType(TaskHandle t, const ProgramType type);
		assert measure_type in self.prog_type, "Invalid measurement type!"
		measurement_type_code = self.prog_type[measure_type]
		ret = self.dll.setMeasurementType(self.task_handle, measurement_type_code)
		self.__check_ret(ret)

	###############################################################

	# TODO: Also UNTESTED!
	def setGainSetting(self, gain_setting):
		'''
		Control IF gain.

		The AVMU has the facility for configurable  gain within the IF stage
		of the RF pipeline.

		Args:
			gain_setting (str) A value in the IfGain Enum (e.g. "AVMU_GAIN_*")

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the task is either UNINITIALIZED or STOPPED.

		'''
		self.log.debug("setGainSetting call")
		# ErrCode getIfGain(TaskHandle t, IfGain* new_gain);
		assert gain_setting in self.if_gain_settings, "Invalid gain value!"
		new_if_gain_enum = self.if_gain_settings[gain_setting]
		print("Specified gain: %s, %s" % (gain_setting, new_if_gain_enum))
		ret = self.dll.setIfGain(self.task_handle, new_if_gain_enum)
		self.__check_ret(ret)


	def getGainSetting(self):
		'''
		Get the current IF gain setting.

		Returns:
			(str) Current IF gain settings.
		'''
		self.log.debug("getGainSetting call")
		# ErrCode setIfGain(TaskHandle t, IfGain new_gain);
		current_gain_setting = self.ffi.new("int *", 0)
		ret = self.dll.getIfGain(self.task_handle, current_gain_setting)
		self.__check_ret(ret)

		if current_gain_setting[0] in self.if_gain_inverse:
			return self.if_gain_inverse[current_gain_setting[0]]
		else:
			raise avmu_exceptions.Avmu_Exception("Failed to decode returned gain settings (%s)!" % current_gain_setting[0])


	def setReceiver12dBPad(self, insert_pad):
		'''
		Control whether an optional 12dB pad is inserted in the internal RF chain.


		Args:
			insert_pad (bool): Whether the pad is inserted.
		'''
		self.log.debug("setReceiver12dBPad call")
		# ErrCode getIfGain(TaskHandle t, IfGain* new_gain);
		assert isinstance(insert_pad, bool), "insert_pad must be a boolean!"

		ret = self.dll.setReceiver12dBPad(self.task_handle, insert_pad)
		self.__check_ret(ret)

	def getReceiver12dBPad(self):
		'''
		TODO: Untested!

		Get whether the optional 12DB pad is inserted into the RF chain.

		Returns:
			(bool) True if pad is enabled.
		'''
		self.log.debug("getReceiver12dBPad call")
		# ErrCode setIfGain(TaskHandle t, IfGain new_gain);
		is_inserted = self.ffi.new("bool *", False)
		ret = self.dll.getReceiver12dBPad(self.task_handle, is_inserted)
		self.__check_ret(ret)

		return is_inserted[0]



	###############################################################

	def getNumberOfFrequencies(self):
		'''

		Get the number of frequency points for the sweep configured for
		the Task. If no frequencies have been set, this function returns 0.

		Returns:
			(int) Number of frequency points in the current sweep.

		'''
		self.log.debug("getNumberOfFrequencies call")
		# Signature: unsigned int getNumberOfFrequencies(TaskHandle t);
		freqNum = self.dll.getNumberOfFrequencies(self.task_handle)
		return freqNum


	def utilGenerateLinearSweep(self, startF_mhz, stopF_mhz, points):
		'''

		Generates a linear sweep with the requested parameters. Note that the start and end
		frequency will be adjusted as documented in utilFixLinearSweepLimits() so that all
		frequency points fall on exactly generateable values. This function internally calls
		setFrequencies() with the resulting array. The caller can retrieve the frequency list
		with the getFrequencies() function. Since it changes the frequencies this function
		is only available in the TASK_STOPPED state.

		If `startFreq` == `endFreq`, the hardware will effectively be placed in zero-span
		mode, as it will repeatedly sample the same frequency for the duration of the
		sweep. This is a valid operating mode.

		Args:
			startF_mhz (float) Start frequency of sweep in Mhz
			stopF_mhz (float) End frequency of sweep in Mhz
			N (int) Number of points to sample

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Too_Many_Points` if N is larger than \
			                                the maximum allowed (see HardwareDetails)
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Freq_Out_Of_Bounds` if a frequency is \
			                                beyond the allowed min/max. (You can get the min and max from \
			                                the HardwareDetails struct returned by :func:`getHardwareDetails()`)
		'''
		self.log.debug("utilGenerateLinearSweep call")
		# Signature: ErrCode utilGenerateLinearSweep(TaskHandle t, const double startFreq, const double endFreq, const unsigned int N);
		ret = self.dll.utilGenerateLinearSweep(self.task_handle, startF_mhz, stopF_mhz, points)
		self.__check_ret(ret)

	###############################################################

	def getTimeout(self):
		'''
		Get the current UDP socket transport timeout.

		Returns:
			(int) socket timeout, in milliseconds.
		'''
		self.log.debug("getTimeout call")
		# Signature: unsigned int getTimeout(TaskHandle t);
		timeout = self.dll.getTimeout(self.task_handle)
		self.log.debug("Current timeout value: %s ms", timeout)
		return timeout

	def setTimeout(self, timeout_ms):
		'''
		Sets the default time to wait, in milliseconds, for a unit to reply to a command
		before giving up and returning an ERR_NO_RESPONSE condition. For the measurement
		functions, this is the amount of time to wait beyond the expected sweep time.
		When a Task is created, the timeout value defaults to 100.

		A timeout value of 0 results in non-blocking call, where the call will return
		immediately if there is no data in the OS receive buffer.

		Args:
			timeout_ms (int) - Requested timeout in milliseconds.

		Returns:
			Nothing
		'''
		self.log.debug("setTimeout call")
		# Signature: ErrCode setTimeout(TaskHandle t, const unsigned int timeout);
		self.log.debug("Setting socket timeout to: %s ms", timeout_ms)
		ret = self.dll.setTimeout(self.task_handle, timeout_ms)
		self.__check_ret(ret)

	###############################################################

	def isSerialPortPresent(self):
		'''
		Get whether the remote hardware has a serial port peripheral.

		Returns:
			True if the hardware supports a serial port, False if not.
		'''
		self.log.debug("isSerialPortPresent call")
		# Signature: ErrCode isSerialPortPresent(TaskHandle t, bool* present);
		present = self.ffi.new("bool *", False)
		ret = self.dll.isSerialPortPresent(self.task_handle, present)
		self.__check_ret(ret)
		return present[0]

	def setSerialPortFeature(self, enable, buffer_size = 128):
		'''
		If the remote hardware has a serial port, this call configures it.

		Args:
			enable (bool) Enables/Disables the reception of serial data by the AVMU.
			buffer_size (int) Size of the serial RX buffer to allocate. This can be \
			    any value > 1 and < 255.


		raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Feature_Not_Present`:   if the remote AVMU \
			    does not have a serial port.

		'''
		self.log.debug("setSerialPortFeature call")
		# Signature: ErrCode setSerialPortFeature(TaskHandle t, const bool enable, const unsigned int buffer_size);
		assert (buffer_size < 256)
		self.serial_buf_sz = buffer_size
		ret = self.dll.setSerialPortFeature(self.task_handle, enable, buffer_size)
		self.__check_ret(ret)

	###############################################################

	def isShaftEncoderPresent(self):
		'''
		Get whether the AVMU has support for external shaft encoder inputs.

		Returns:
			(bool) whether the connected AVMU has support for encoders

		'''
		self.log.debug("isShaftEncoderPresent call")
		# Signature: ErrCode isShaftEncoderPresent(TaskHandle t, bool* present);
		present = self.ffi.new("bool *", False)
		ret = self.dll.isShaftEncoderPresent(self.task_handle, present)
		self.__check_ret(ret)
		return present[0]

	def setShaftEncoderFeature(self, enable, resetOnStart=True):
		'''
		If the AVMU supports encoders, this allows you to enable the encoders,
		and optionally reset the cumulative encoder counter.

		Args:
			enable(bool) Tell the hardware to sample the encoders.
			resetOnStart(bool) Reset the accumulated encoder counter in the hardware \
			             on sweep start. If this is not set, the value accumulated from \
			             any previous motion will remain in the accumulator, and be \
			             present when the hardware is started.


		raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Feature_Not_Present`:   if the remote AVMU \
			    does not have encoder support.


		'''
		self.log.debug("setShaftEncoderFeature call")
		# Signature: ErrCode setShaftEncoderFeature(TaskHandle t, const bool enable);
		ret = self.dll.setShaftEncoderFeature(self.task_handle, enable, resetOnStart)
		self.__check_ret(ret)

	###############################################################

	# Note: UNTESTED!
	def configureTddSettings(self,
				tddEnabled        = False,
				nullingEnabled    = False,
				powerAmpState     = False,
				slave             = False,
				attenuatorEnabled = False,
				lnaEnabled        = False,
				attenuatorValue   = 0,
				tx                = 0,
				tx_to_rx1         = 0,
				rx1               = 0,
				rx1_to_rx2        = 0,
				rx2               = 0,
				rx2_to_tx         = 0,
			):
		self.log.debug("configureTddSettings call")
		# Signature: ErrCode setShaftEncoderFeature(TaskHandle t, const bool enable);
		ret = self.dll.configureTddSettings(
			self.task_handle,
			True,
			bool(tddEnabled),
			bool(nullingEnabled),
			bool(powerAmpState),
			bool(slave),
			bool(attenuatorEnabled),
			bool(lnaEnabled),
			bool(attenuatorValue),
			int(tx),
			int(tx_to_rx1),
			int(rx1),
			int(rx1_to_rx2),
			int(rx2),
			int(rx2_to_tx))
		self.__check_ret(ret)

	###############################################################

	def getState(self):

		'''
		 Get the current state of the Task.

		Returns:
			(str) One of 'TASK_RUNNING', 'TASK_STARTED', 'TASK_STOPPED' or 'TASK_UNINITIALIZED'.

		'''
		self.log.debug("getState call")
		# Signature: TaskState getState(TaskHandle t);
		state = self.dll.getState(self.task_handle)
		for state_name, state_val in self.run_state.items():
			if state_val == state:
				return state_name

		raise avmu_exceptions.Avmu_Exception_Wrong_State("State value returned is not known (%s)!" % state)

	#################################################################################
	#        Diagnostics
	#################################################################################

	def getHardwareDetails(self):
		'''
		Get the hardware details for the unit associated with this AvmuInterface instance.

		TODO: VALIDATE THIS!

		Args:
			None

		Returns:
			A dictionary of the HardwareDetails members -> value mappings.  \
			If the Task has not yet been initialized, the returned dict has all values set to 0.

		'''
		self.log.debug("getHardwareDetails call")
		# Signature: HardwareDetails getHardwareDetails(TaskHandle t);
		hardwareDetails = self.dll.getHardwareDetails(self.task_handle)

		ret = {
			"minimum_frequency" : hardwareDetails.minimum_frequency,
			"maximum_frequency" : hardwareDetails.maximum_frequency,
			"maximum_points"    : hardwareDetails.maximum_points,
			"serial_number"     : hardwareDetails.serial_number,

			"band_boundaries"   : [hardwareDetails.band_boundaries[x]
										for x in range(hardwareDetails.number_of_band_boundaries)],
			"switch_board_type" : hardwareDetails.swbd_type,
			"feature_flags"     : {

				"has_encoders"           : hardwareDetails.hardware_features.has_encoders,
				"has_serial_port"        : hardwareDetails.hardware_features.has_serial_port,
				"has_attenuators"        : hardwareDetails.hardware_features.has_attenuators,
				"has_multiple_receivers" : hardwareDetails.hardware_features.has_multiple_receivers,
				"has_scan_trigger_in"    : hardwareDetails.hardware_features.has_scan_trigger_in,
				"has_scan_trigger_out"   : hardwareDetails.hardware_features.has_scan_trigger_out,

			}
		}
		return ret

	def utilPingUnit(self, tries = 5):
		'''
		Sends an "are you there" message to the unit.

		Note that this function should not be
		called while a frequency sweep is ongoing, because it causes that sweep to prematurely
		halt and respond to this message instead. This is only an issue in multithreaded code or
		when in async mode, as the sync data acquisition functions are blocking. This function
		waits for a reply for the length of time specified by getTimeout() before giving up.

		Note that this can be called from any state, provided an IP and port are present.

		Raises an exception if the unit did not respond, returns nothing otherwise.

		Args:
			tries (int) Number of pings to send before concluding a unit is not responding.
			      The call will block for at most ``tries`` * ``timeout`` milliseconds.
			      Note that certain states can cause the hardware to be unable to
			      respond to the first ping, so tries > 1 is generally recommended.

		Raises:

			:class:`~avmu.avmu_exceptions.Avmu_Exception_Socket` if there was a problem sending a message
			:class:`~avmu.avmu_exceptions.Avmu_Exception_No_Response` if the unit did not respond to commands
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Missing_Ip` if no IP address has been set
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Missing_Port` if no port has been set

		'''
		self.log.debug("utilPingUnit call")
		# Signature: ErrCode utilPingUnit(TaskHandle t);
		ret = self.dll.utilPingUnit(self.task_handle, tries)
		self.__check_ret(ret)
		self.log.debug("utilPingUnit return")
		return True

	def versionString(self):
		'''
		Returns a string describing the version of the DLL and its components.

		Returns:
			String describing the AVMU DLL components and version numbers.
		'''
		self.log.debug("versionString call")
		# Signature: const char* versionString();
		ret = self.dll.versionString()
		return self.ffi.string(ret).decode("ascii")

	#################################################################################
	#        Utilities
	#################################################################################

	def utilFixLinearSweepLimits(self, startF, endF, npts):
		'''
		Adjusts the start and end of a requested linear sweep with N points such that all
		frequencies in the sweep will land on exactly generateable values, and the inter-point
		spacing is constant across the entire scan. Unequal spacing will cause Doppler noise in
		the data.

		This may move the start and end frequencies of the scan slightly (<1 MHz).

		If the input frequencies are equal, or N is 1, the frequencies are each
		simply adjusted to exactly generateable values.

		TODO: VALIDATE THIS!

		Args:
			startF(float) - Target start frequency for sweep, in Mhz.
			endF(float) - Target end frequency for sweep, in Mhz.
			npts(int) - Number of points to sample, spaced linearly between startFreq and endFreq

		Returns:
			Adjusted start and stop frequencies as a 2-tuple


		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Too_Many_Points` if N is larger than \
			                                the maximum allowed (see HardwareDetails)
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Freq_Out_Of_Bounds` if a frequency is \
			                                beyond the allowed min/max. (You can get the min and max from \
			                                the HardwareDetails struct returned by :func:`getHardwareDetails()`)

		'''
		self.log.debug("utilFixLinearSweepLimits call")
		# Signature: ErrCode utilFixLinearSweepLimits(TaskHandle t, double* startFreq, double* endFreq, const unsigned int N);
		startF = self.ffi.new("double *", startF)
		endF   = self.ffi.new("double *", endF)
		ret = self.dll.utilFixLinearSweepLimits(self.task_handle, startF, endF, npts)
		self.__check_ret(ret)
		return startF[0], endF[0]

	def utilNearestLegalFreq(self, freq):
		'''
		Adjusts a requested frequency, in MHz, to the nearest able to be generated by the
		hardware. This is not available in the TASK_UNINITIALIZED state, as it requires knowledge
		of the hardware not read from the AVMU until initialized.

		Args:
			target_freq (float) - Target frequency to be adjusted

		Returns:
			Adjusted `target_freq` Frequency value as a float

			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Freq_Out_Of_Bounds` if a frequency is \
			                                beyond the allowed min/max. (You can get the min and max from \
			                                the HardwareDetails struct returned by :func:`getHardwareDetails()`)

		'''
		self.log.debug("utilNearestLegalFreq call")
		# Signature: ErrCode utilNearestLegalFreq(TaskHandle t, double* freq);
		freq   = self.ffi.new("double *", freq)
		ret = self.dll.utilNearestLegalFreq(self.task_handle, freq)
		self.__check_ret(ret)
		return freq[0]

	#################################################################################
	#        Execution
	#################################################################################

	def initialize(self):
		'''
		Attempts to talk to the unit specified by the Task's IP address and port, and download
		its details. If it succeeds the Task enters the TASK_STOPPED state.

		Raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Socket` if there was a problem sending a message
			:class:`~avmu.avmu_exceptions.Avmu_Exception_No_Response` if the unit did not respond to commands
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Missing_Ip` if no IP address has been set
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Missing_Port` if no port has been set
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State`  if the Task is not in the TASK_UNINITIALIZED state
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Prom`  if the unit returned hardware details that this DLL doesn't understand


		'''
		self.log.debug("initialize call")
		# Signature: ErrCode initialize(TaskHandle t, progress_callback callback, void* user);
		self.log.debug("Initializing remote device.")
		ret = self.dll.initialize(self.task_handle, self.ffi.NULL, self.ffi.NULL)
		self.__check_ret(ret)
		self.log.debug("Remote device initialized.")


	def beginAsync(self):
		'''
		Begin async data collection.
		This call emits the trigger message that will cause the remote hardware to
		immediately begin acquiring frequency data continuously.

		Note that once beginAsync() has been called, you MUST then call measure()
		periodically so that the UDP recieve buffer will not overflow.
		'''
		self.log.debug("beginAsync call")
		# Signature: ErrCode beginAsync(TaskHandle t);
		ret = self.dll.beginAsync(self.task_handle)
		self.__check_ret(ret)

	def broadcastBeginCommand(self, handles=[]):
		'''

		Broadcast a "begin" command to all avmus on the local network. This will start the
		sweep of every avmu on the network at the current position of its internal PC, and
		therefore depends on the avmus having been preconfigured to the proper start addresses
		using beginAsync().

		Args:
			handles (list of AvmuInterface) The list of AVMU AvmuInterface instances to start \
			        in a coordinated manner. Note that if you have all AVMUs in SYNC_IGNORE mode \
			        this will operate fine, but the AVMUs will not be properly synchronized.

		Raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if any of the the tasks are not in the TASK_RUNNING state.
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Task_Array_Invalid` if the passed list is empty (handle_count == 0)

		'''
		self.log.debug("broadcastBeginCommand call")
		# Signature: ErrCode beginAsync(TaskHandle t);
		handles = [other.__getRawTaskHandle() for other in handles]
		# print("Handles:", handles)
		ret = self.dll.broadcastBeginCommand(handles, len(handles))
		self.__check_ret(ret)

	def haltAsync(self):
		'''
		This causes an immediate halt of the async sweep acquisition in the
		remote hardware.

		Any pending data in the UDP recieve buffer will be flushed, and the
		hardware will be returned to the "started" state.

		To resume async data collection after calling haltAsync(),
		simply call beginAsync().
		'''
		self.log.debug("haltAsync call")
		# Signature: ErrCode haltAsync(TaskHandle t);
		ret = self.dll.haltAsync(self.task_handle)
		self.__check_ret(ret)

	def interruptMeasurement(self):
		'''
		Interrupts one of the measurement functions while it is waiting for data.

		Since the measurement functions are blocking in SYNC mode, this function
		must be called from a different thread. This function returns immediately,
		however the measurement function may continue to block for a short additional
		amount of time.

		Raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STARTED state
		'''
		self.log.debug("interruptMeasurement call")
		# Signature: ErrCode interruptMeasurement(TaskHandle t);
		ret = self.dll.interruptMeasurement(self.task_handle)
		self.__check_ret(ret)

	def start(self):
		'''
		Attempts to program the AVMU using the settings stored in the Task object. If it
		succeeds the Task enters the TASK_STARTED state.


		Raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State`  if the Task is not in the TASK_STOPPED state
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Socket` if there was a problem sending a message
			:class:`~avmu.avmu_exceptions.Avmu_Exception_No_Response` if the unit did not respond to commands
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Prom`  if the unit returned hardware details that this DLL doesn't understand
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Hop` if the hop rate has not yet been specified
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Missing_Freqs` if the frequencies have not yet been specified
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Prog_Overflow` if the size of the program is too large for the hardware's memory \
				(this can happen if there are too many frequencies)
		'''
		self.log.debug("start call")
		# Signature: ErrCode start(TaskHandle t);
		self.log.info("Starting task.")
		ret = self.dll.start(self.task_handle)
		self.__check_ret(ret)


	def stop(self):
		'''
		 Puts the Task object into the TASK_STOPPED state.

		Raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State`  if the Task is not in the TASK_STARTED state

		'''
		self.log.debug("stop call")
		# Signature: ErrCode stop(TaskHandle t);
		self.log.info("Stopping task.")
		ret = self.dll.stop(self.task_handle)
		self.__check_ret(ret)



	def measure(self):
		'''
		Take a measurement.

		In PROG_SYNC mode, this triggers a measurement, and blocks until its response has
		been fully received and decoded.

		This function does various things depending on the sweep trigger type and the program type.
		Sync mode, internal trigger: run avmu program, block until data is received or timeout
		Sync mode, external trigger: block until data is received or timeout
		Async mode, internal trigger: block until 1 program's worth of data is received or timeout
		Async mode, external trigger: block until 1 program's worth of data is received or timeout

		Note that beginAsync() must be called prior to measure() in async mode.
		The pattern is beginAsync(), measure(), measure(), measure(), ..., haltAsync()

		If the AvmuTask is in async mode (setMeasurementType called with PROG_ASYNC)
		this simply serves to consume any UDP messages generated by the hardware. Note that
		since the hardware asynchronously emits data continuously, you therefore
		*must* call this periodically, or you risk overrunning the socket receive buffer.

		Typically, overrunning the buffer will result in a ``Avmu_Exception_Bytes`` return code, but
		it's theoretically possible to have an entire sweep be dropped (either if it
		fits in a single packet, or if all the packets for the sweep get dropped).

		In this case, the only indicator of data loss will be a non-monotonic sweep
		number in the returned data.

		In general, it's probably ideal to just continualy call this.

		In both cases, measure() places the acquired data in the RX queue, ready
		to be extracted by calling extractSweepData().

		Raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the AvmuInterface is in PROG_SYNC mode, and the AvmuInterface \
		                         state is not TASK_STARTED.
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the AvmuInterface is in PROG_ASYNC mode, and the AvmuInterface \
		                         state is not TASK_RUNNING.

			:class:`~avmu.avmu_exceptions.Avmu_Exception_Bytes` if the received data is corrupted in some manner, or incomplete.
			:class:`~avmu.avmu_exceptions.Avmu_Exception_No_Measured_Paths` if there are no paths added to measure.

		Rare ``Avmu_Exception_Bytes``, particularly when not on a dedicated subnet are
		not entirely unknown, but they shouldn't happen frequently.

		If you encounter large numbers of ``Avmu_Exception_Bytes`` returns, there can be a number of
		causes. TCP Checksum offloading has been known to cause this, as could
		not calling measure() frequently enough. Lastly, check the intervening network
		hardware. You might have a failing switch/hub.

		Note that the AVMU hardware is particularly intolerant of network hubs,
		as its UDP transport does not support data retransmits. Any collisions
		will result in a corrupted sweep (but really, it's 2018+, why are you
		running a network hub?).

		'''
		self.log.debug("measure call")
		# Signature: ErrCode measure(TaskHandle t);
		ret = self.dll.measure(self.task_handle)
		self.__check_ret(ret)


	def getnumberOfEnabledReceivers(self):
		'''
		Returns the number of receivers in the AVMU.
		'''
		self.log.debug("getnumberOfEnabledReceivers call")
		# ErrCode getnumberOfEnabledReceivers(TaskHandle t, int*  num_receivers_enabled);
		present = self.ffi.new("int *", 0)
		ret = self.dll.getnumberOfEnabledReceivers(self.task_handle, present)
		self.__check_ret(ret)
		return present[0]

	def __decodeEnabledReceivers(self, enable_bitmap):
		self.log.debug("__decodeEnabledReceivers call")
		ret = []
		for x in range(8):
			if enable_bitmap & 1 << x:
				ret.append(x)
		return ret

	def getEnabledReceivers(self):
		'''
		Returns a list describing which of the AVMU's receivers will return data.
		'''
		self.log.debug("getEnabledReceivers call")
		# ErrCode getEnabledReceivers(TaskHandle t,         char* enabled_receivers_mask);
		# Use an int8_t because it casts to char cleanly, but
		# doesn't try to act like a bytes character everywhere.
		enable_mask = self.ffi.new("int8_t *", 0)
		ret = self.dll.getEnabledReceivers(self.task_handle, self.ffi.cast("char *", enable_mask))
		self.__check_ret(ret)

		self.active_receivers = self.__decodeEnabledReceivers(enable_mask[0])
		return self.active_receivers

	def __setEnabledReceivers(self, enable_mask):
		self.log.debug("__setEnabledReceivers call")
		# ErrCode setEnabledReceivers(TaskHandle t,         char  enabled_receivers_mask);
		ret = self.dll.setEnabledReceivers(self.task_handle, chr(enable_mask).encode("ascii"))
		self.__check_ret(ret)

		# Query and update the enabled recievers config
		self.getEnabledReceivers()

	def setEnabledReceivers(self, enable_list):
		'''
		Configure which of the receivers in the AVMU will return data.
		'''
		self.log.debug("setEnabledReceivers call")
		valid_receivers = [0, 1, 2, 3]
		assert(all([tmp in valid_receivers for tmp in enable_list]))

		mask = 0
		for item in enable_list:
			mask |= 1 << item
		self.__setEnabledReceivers(mask)

	def __extract_sweep_data_int(self, tx_p_enum, rx_p_enum):
		self.log.debug("__extract_sweep_data_int call")

		point_num = self.getNumberOfFrequencies()
		recs      = self.getEnabledReceivers()

		# print("getEnabledReceivers: ", self.getEnabledReceivers())

		sdat_struct = self.ffi.new("SweepDataStruct *")
		sdat_struct.serial_data_bytes = self.ffi.new("unsigned char [{size}]".format(size=self.serial_buf_sz))

		receiver_arrs = [
				(np.zeros(point_num, dtype=np.float64), np.zeros(point_num, dtype=np.float64))
			for
				dummy
			in
				range(len(recs))
		]

		i_initializer = [self.ffi.cast("double *", tup[0].ctypes.data) for tup in receiver_arrs]
		q_initializer = [self.ffi.cast("double *", tup[1].ctypes.data) for tup in receiver_arrs]

		iarr = self.ffi.new("double*[]", i_initializer)
		qarr = self.ffi.new("double*[]", q_initializer)

		sdat_struct.points.I = iarr
		sdat_struct.points.Q = qarr

		ret = self.dll.extractSweepData(self.task_handle, sdat_struct, tx_p_enum, rx_p_enum)
		self.__check_ret(ret)

		# There's an extra copy here I don't like. Not sure how to work around it, though.

		result = {}
		for x in range(len(recs)):
			result[recs[x]] =np.empty(point_num, dtype=np.complex128)

			result[recs[x]].real = receiver_arrs[x][0]
			result[recs[x]].imag = receiver_arrs[x][1]

		# print("ExtractAllPaths I mean: {}, Q mean: {}".format(np.mean(i), np.mean(q)))

		result_meta = {
			'avmu_ip' : self.getIPAddress(),
			'tx_port'   : self.tx_paths_enum_int[tx_p_enum],
			'tx_port_s' : self.tx_paths_enum_str[tx_p_enum],
			'rx_port'   : self.rx_paths_enum_int[rx_p_enum],
			'rx_port_s' : self.rx_paths_enum_str[rx_p_enum],

			'timestamp_ticks'     : sdat_struct.timestamp_ticks,
			'timestamp_seconds'   : sdat_struct.timestamp_seconds,
			'sweep_number'        : sdat_struct.sweep_number,

			'shaft_encoder_left'  : sdat_struct.shaft_encoder_left,
			'shaft_encoder_right' : sdat_struct.shaft_encoder_right,

			'serial_data_age'   : sdat_struct.serial_data_age,
			'serial_data_bytes' : bytes([sdat_struct.serial_data_bytes[x] for x in range(self.serial_buf_sz)]),

		}
		return result, result_meta



	def __extract_sweep_data_int_old(self, tx_p_enum, rx_p_enum, did_transmit):
		self.log.debug("__extract_sweep_data_int_old call")
		# Signature: ErrCode extractSweepData(TaskHandle t, RFPath path, SweepDataStruct* data);

		point_num = self.getNumberOfFrequencies()
		sdat_struct = self.ffi.new("SweepDataStruct *")
		sdat_struct.serial_data_bytes = self.ffi.new("unsigned char [{size}]".format(size=self.serial_buf_sz))

		i = np.empty(point_num, dtype=np.float64)
		q = np.empty(point_num, dtype=np.float64)


		sdat_struct.points.I = self.ffi.cast("double *", i.ctypes.data)
		sdat_struct.points.Q = self.ffi.cast("double *", q.ctypes.data)

		ret = self.dll.extractSweepData(self.task_handle, sdat_struct, tx_p_enum, rx_p_enum, did_transmit)
		self.__check_ret(ret)

		result = np.empty(point_num, dtype=np.complex128)

		# There's an extra copy here I don't like. Not sure how to work around it, though.
		result.real = i
		result.imag = q
		result_meta = {
			'tx_port'   : self.tx_paths_enum_int[tx_p_enum],
			'tx_port_s' : self.tx_paths_enum_str[tx_p_enum],
			'rx_port'   : self.rx_paths_enum_int[rx_p_enum],
			'rx_port_s' : self.rx_paths_enum_str[rx_p_enum],

			'timestamp_ticks'     : sdat_struct.timestamp_ticks,
			'timestamp_seconds'   : sdat_struct.timestamp_seconds,
			'sweep_number'        : sdat_struct.sweep_number,

			'shaft_encoder_left'  : sdat_struct.shaft_encoder_left,
			'shaft_encoder_right' : sdat_struct.shaft_encoder_right,

			'serial_data_age'   : sdat_struct.serial_data_age,
			'serial_data_bytes' : bytes([sdat_struct.serial_data_bytes[x] for x in range(self.serial_buf_sz)]),

		}
		return result, result_meta


	def extractSweepData(self, tx_path, rx_path):
		'''
		Extract the sweep data for a given path described by a TX port and RX port.

		Note that this interface has the limitation that it does not support multiple measurements
		of the same path within the overall AVMU acqusision frame.

		Args:
			tx_path (str): TX port of the path to extract.
			rx_path (str): RX port of the path to extract.

		Returns:
			- A 2-tuple of containing (data, metadata)
			    data is a numpy complex array with the dimensions of 1 x ``getNumberOfFrequencies()``,
			      where each value bin corresponds to the associated frequency bin in ``getFrequencies()``.
			    metadata is a dict containing the sweep metadata associated with the sweep ``data``.
			      It contains the following key-> value pairs:
				- ``tx_port`` - Transmitting port, as a integer.
				- ``tx_port_s`` - Transmitting port, as a string compatible with ``addPathToMeasure()``.
				- ``rx_port`` - Receiving port, as a integer.
				- ``rx_port_s`` - Receiving port, as a string compatible with ``addPathToMeasure()``.
				- ``timestamp_ticks``   Currently unused. Read as zero.
				- ``timestamp_seconds`` Currently unused. Read as zero.
				- ``sweep_number``      Monotonically increasing counter of the number of sweeps (note:
					sweeps, not frames)
				- ``shaft_encoder_left``  Shaft encoder accumulator for the left-hand shaft encoder.
				    If not enabled, will return 0.
				- ``shaft_encoder_right`` Shaft encoder accumulator for the right-hand shaft encoder.
				    If not enabled, will return 0.
				- ``serial_data_age`` Age of the serial port data. If serial is not enabled, will return 0.
				- ``serial_data_bytes`` Serial data itself, as a bytes array. If not enabled, will be an empty bytes object.

		raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Bad_Path`:   if a path value specified is invalid.

		'''
		self.log.debug("extractSweepData call")
		# Signature: ErrCode extractSweepData(TaskHandle t, RFPath path, SweepDataStruct* data);
		assert tx_path in self.tx_paths, "Invalid path!"
		assert rx_path in self.rx_paths, "Invalid path!"
		tx_p = self.tx_paths[tx_path]
		rx_p = self.rx_paths[rx_path]

		return self.__extract_sweep_data_int(tx_p, rx_p)

	def __extractSweepDataIntPath(self, tx_path_int, rx_path_int):
		self.log.debug("__extractSweepDataIntPath call")
		# Signature: ErrCode extractSweepData(TaskHandle t, RFPath path, SweepDataStruct* data);
		assert tx_path_int in self.tx_paths_int_enum, "Invalid path!"
		assert rx_path_int in self.rx_paths_int_enum, "Invalid path!"
		tx_p = self.tx_paths_int_enum[tx_path_int]
		rx_p = self.rx_paths_int_enum[rx_path_int]

		return self.__extract_sweep_data_int(tx_p, rx_p)

	def extractAllPaths(self):
		'''
		Given a set of measured paths that are unique, extract every path and return them
		as a list.

		This call presumes the user has called ``measure()`` prior to it's invocation.
		If not, it will raise :class:`~avmu.avmu_exceptions.Avmu_Exception_Path_Has_No_Data`:   if you didn't call ``measure()``

		Returns:
			A list of 2-tuples ``(path_info, data)``.

			``path_info`` is a dict containing the following:

				- ``who_is_transmitting``  Which unit is doing the transmitting. This is
				  only relevant in a multi-AVMU context. Single-AVMU setups can ignore it.
				- ``port_is_transmitting``   Which port on the transmitting unit is active. This is
				  only relevant in a multi-AVMU context. Single-AVMU setups can ignore it.
				- ``tx_path``              Which port on the current AVMU is transmitting. For multi AVMU
				  configurations, when the current AVMU is not transmitting, this will be TX_NONE.
				- ``rx_path``              Which port on the current AVMU is receiving.
				- ``receiver_channel``     Which receiver is the dataset for. At the moment, this
				  is ALWAYS 0, as no multi-receiver AVMUs are available.


				In the general single-unit case, you can ignore all the keys other then ``tx_path`` and ``rx_path``.

			``data`` is a dict containing two keys:

				- ``data`` The actual sweep's data. This is directly the ``data`` return component from
				  the ``extractSweepData()`` call.
				- ``meta`` The sweep's metadata. This is directly the ``metadata`` return component from
				  the ``extractSweepData()`` call.

		raises:
			:class:`~avmu.avmu_exceptions.Avmu_Exception_Path_Has_No_Data`:   if you didn't call ``measure()``

		'''
		self.log.debug("extractAllPaths call")
		# Python-only convenience function.
		ret = []
		# print("extractAllPaths() for radar ", self.getIPAddress())
		for who_is_transmitting, port_is_transmitting, tx_path, rx_path in self.measured_paths:
			# print("Extracting path: ", tx_path, rx_path, transmitted)
			rx_dict, meta = self.__extractSweepDataIntPath(tx_path, rx_path)
			for receiver_channel, data in rx_dict.items():
				ret.append((
					{
						'who_is_transmitting'  : who_is_transmitting,
						'port_is_transmitting' : port_is_transmitting,
						'tx_path'              : tx_path,
						'rx_path'              : rx_path,
						'receiver_channel'     : receiver_channel,
					},
					{
						'data' : data,
						'meta' : meta
					}
					))
		return ret



	def setSyncPulseMode(self, sync_mode):
		'''
		Configure the inter-AVMU synchronization pulse mode for the specified AVMU.

		The SyncPulseMode is used to allow coordinated acquisition between
		multiple AVMUs. In coordinated operation, the master unit is set to
		SYNC_GENERATE, and any slave units are set to SYNC_RECEIVE. This
		causes the master to emit sweep start outputs, which are consumed
		by the slave units.

		Note that this requires specialized hardware support, and dedicated
		syncronization wiring.

		Additionally, to properly synchronize, all units must receive their
		start command within 1 ms of each other, which is accomplished via
		the broadcastBeginCommand() call. This also requires the
		shared-communication-infrasturcture provided by the createSharedTask()
		call. All AVMU tasks used in a coordinated acqusition must share
		their internal communication object.

		Args:
			sync_mode (str) One of "SYNC_IGNORE", "SYNC_GENERATE" or "SYNC_RECEIVE"

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State`if the Task is not in the ``TASK_UNINITIALIZED`` or ``TASK_STOPPED`` state

		'''
		self.log.debug("setSyncPulseMode call")
		assert sync_mode in self.sync_pulse_mode

		mode = self.sync_pulse_mode[sync_mode]
		ret = self.dll.setSyncPulseMode(self.task_handle, mode)
		self.__check_ret(ret)

	def getSyncPulseMode(self):
		'''
		Get the current sync pulse mode for this unit.

		Returns:
			Current sync-pulse mode as a string.
		'''

		self.log.debug("getSyncPulseMode call")
		# ErrCode getnumberOfEnabledReceivers(TaskHandle t, int*  num_receivers_enabled);
		sync_mode = self.ffi.new("int *", 0)
		ret = self.dll.getSyncPulseMode(self.task_handle, sync_mode)
		self.__check_ret(ret)

		for key, value in self.sync_pulse_mode.items():
			if value == sync_mode[0]:
				return key

		raise avmu_exceptions.Avmu_Exception("Failed to decode returned sync pulse mode (%s)!" % sync_mode[0])


	def addExclusionBand(self, start_freq, stop_freq):
		'''
		Exclusion bands are used to prevent the AVMU from transmitting in specific
		frequency ranges. This is generally used to "mask out" sensitive RF regions,
		so things like GPS are not negatively affected by the AVMU.

		The AVMU will still take data at points within the exclusion band, but
		the RF output will be disabled.

		Multiple exclusion bands are logically ORed together. Basically, if
		a frequency point falls within *any* exclusion band, it will result in the
		RF output being disabled.

		As such, repeated or overlapping exclusion bands are valid (though
		they are somewhat pointless).

		Note that you cannot add exclusion bands before connecting to an AVMU.

		Note that start_freq must be smaller then stop_freq.

		Args:
			start_freq (float, MHz) - Start of the exclusion band
			stop_freq (float, MHz) - End of the exclusion band (MUST be larger then start_freq)

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state

		'''
		self.log.debug("addExclusionBand call")
		# ErrCode addExclusionBand(TaskHandle t, double start_freq, double stop_freq);
		assert stop_freq > start_freq, "The stop frequency must be larger then the start frequency"
		ret = self.dll.addExclusionBand(self.task_handle, start_freq, stop_freq)
		self.__check_ret(ret)

	def clearExclusionBands(self):
		'''
		Clear the internal list of exclusion bands stored in the AvmuInterface.

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Err_Invalid_Parameter` if the parameters do not make sense.
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state

		'''
		self.log.debug("clearExclusionBands call")
		# ErrCode clearExclusionBands(TaskHandle t);
		ret = self.dll.clearExclusionBands(self.task_handle)
		self.__check_ret(ret)

	def getExclusionBandCount(self):
		'''
		Get the number of separate exclusion bands configured in the AvmuInterface.

		This is required to enumerate the active exclusion bands.

		Returns:
			exclusion_band_count(int) number of exclusion bands configured.

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state

		'''
		self.log.debug("getExclusionBandCount call")
		# ErrCode getExclusionBandCount(TaskHandle t, int* idx);

		exclusion_band_count = self.ffi.new("int *", 0)

		ret = self.dll.getExclusionBandCount(self.task_handle, exclusion_band_count)
		self.__check_ret(ret)

		return exclusion_band_count[0]

	def getExclusionBand(self, band_idx):
		'''
		Get the (start_f_mhz, stop_f_mhz) values for exclusion band band_idx.

		Given a index of 0 <= idx < getExclusionBandCount(), return
		the corresponding exclusion band start/end frequency.

		Return:
			2-tuple of start_f_mhz (double), stop_f_mhz (double)

		Raises:
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Wrong_State` if the Task is not in the TASK_STOPPED state
			 :class:`~avmu.avmu_exceptions.Avmu_Exception_Err_Index_Out_Of_Bounds` if the specified index is not valid.  \
			                                                                      Note that when no exclusion bands are  \
			                                                                      specified, ALL possible values for idx  \
			                                                                      are therefore invalid.
		'''
		self.log.debug("getExclusionBand call")
		# ErrCode getExclusionBand(TaskHandle t, int idx, double* start_freq, double* stop_freq);

		start_f = self.ffi.new("double *", 0)
		stop_f  = self.ffi.new("double *", 0)

		ret = self.dll.getExclusionBand(self.task_handle, band_idx, start_f, stop_f)
		self.__check_ret(ret)

		return start_f[0], stop_f[0]

	def getPreciseTimePerFrame(self):
		'''
		Get the precise time (in seconds) that a Frame will take.

		This function only operates properly when a AVMU is in the started or running states, as
		the sweep program is not computed until the transition to the started state.

		Returns:
			Frame time, as a double, in fractional seconds.
			     If the task is not valid, or in the correct state,
			     the return value is -1.
		'''

		ret = self.dll.getPreciseTimePerFrame(self.task_handle)

		return ret




